/*****************************************************/
/* contents.c                                        */
/* -- Demonstrates providing protect mode "window"   */
/*    into large (>64KB) block of memory.            */
/* -- To build: cc -d -DSTRICT contents.c commdlg.lib*/
/*****************************************************/
#include <windows.h>
#include <windowsx.h>
#include <commdlg.h>
#include <memory.h>
#include <limits.h>

#define midOpen 1000      /* Open file menu item id. */
#define cbBuf   0x00010000          /* 64KB buffers. */

/* Maximum line map size. */
#define clinMax ((cbBuf / sizeof(DWORD)) - 1)

/* Window offsets into for text and line map ptrs. */
#define ibText  0
#define ibMap   sizeof(LPSTR)

const char szClass[] = "ContentsClass";
const char szTitle[] = "Contents";

LRESULT CALLBACK __export LwWndProc(HWND hwnd, UINT wm,
  WPARAM wParam, LPARAM lParam);
void ReadText(HWND hwnd);
void PaintText(HWND hwnd);
void ScrollText(HWND hwnd, int ilin, int wType);

/* Selector functions. */
void  FreeSel(WORD sel);
DWORD IbBaseGetSel(WORD sel);
WORD  SelAlloc(DWORD ibLim);
void  SetBaseSelIb(WORD sel, DWORD ibBase);

WORD SelAlloc(DWORD ibLim)
/*****************************************************/
/* -- Allocate and initialize a LDT selector.        */
/* -- ibLim : Limit to set for selector.             */
/*****************************************************/
    {
    WORD sel;

    _asm mov ax, 0000h;    /* Allocate the selector. */
    _asm mov cx, 0001h;
    _asm int 31h;       /* Should check for success. */
    _asm mov sel, ax;

    if (0x00100000 < ibLim)
        ibLim |= 0x0fff;
    _asm mov ax, 0008h;            /* Set the limit. */
    _asm mov bx, sel;
    _asm mov cx, WORD PTR [ibLim + 2];
    _asm mov dx, WORD PTR [ibLim];
    _asm int 31h;       /* Should check for success. */

    _asm mov ax, 0009h;
    _asm mov bx, sel;      /* Set the access rights. */
    _asm mov cx, 00f3h;
    _asm int 31h;       /* Should check for success. */

    return sel;
    }

void SetBaseSelIb(WORD sel, DWORD ibBase)
/*****************************************************/
/* -- Set the base address for the selector.         */
/*****************************************************/
    {
    _asm mov ax, 0007h;
    _asm mov bx, sel;
    _asm mov cx, WORD PTR [ibBase + 2];
    _asm mov dx, WORD PTR [ibBase];
    _asm int 31h; /* Should check for success. */
    }

DWORD IbBaseGetSel(WORD sel)
/*****************************************************/
/* -- Get the base address for the selector.         */
/*****************************************************/
    {
    DWORD ib;

    _asm mov ax, 0006h;
    _asm mov bx, sel;
    _asm int 31h;      /* Should check for success. */
    _asm mov WORD PTR [ib + 2], cx;
    _asm mov WORD PTR [ib], dx;

    return ib;
    }

void FreeSel(WORD sel)
/*****************************************************/
/* -- Frees the given selector.                      */
/*****************************************************/
    {
    _asm mov ax, 0001h;
    _asm mov bx, sel;
    _asm int 31h;
    }

#ifdef __BORLANDC__
    #pragma argsused
#endif
int PASCAL WinMain(HINSTANCE hins, HINSTANCE hinsPrev,
  LPSTR lpsz, int wShow)
/*****************************************************/
/* -- Entry point.                                   */
/*****************************************************/
    {
    HWND hwnd;
    MSG  msg;

    if (NULL == hinsPrev)      /* Create main class. */
        {
        WNDCLASS wcs;

        wcs.style = CS_HREDRAW | CS_VREDRAW;
        wcs.lpfnWndProc = LwWndProc;
        wcs.cbClsExtra = 0;
        wcs.cbWndExtra =
          sizeof(LPSTR) + sizeof(LPLONG);
        wcs.hInstance = hins;
        wcs.hIcon = LoadIcon(NULL, IDI_APPLICATION);
        wcs.hCursor = LoadCursor(NULL, IDC_ARROW);
        wcs.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
        wcs.lpszMenuName = NULL;
        wcs.lpszClassName = szClass;
        RegisterClass(&wcs);
        }

    hwnd = CreateWindow(szClass, szTitle,
      WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, 
      CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
      NULL, NULL, hins, NULL);

    /* Lazy -- put Open File on system menu. */
    AppendMenu(GetSystemMenu(hwnd, FALSE), MF_STRING,
      midOpen, "&Open...");

    ShowWindow(hwnd, wShow);
    while (GetMessage(&msg, NULL, NULL, 0))
        {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
        }
    return msg.wParam;
    }

void ReadText(HWND hwnd)
/*****************************************************/
/* -- Read the file into memory.                     */
/*****************************************************/
    {
    HFILE        hfil;                /* Input file. */
    OPENFILENAME ofn;  /* Needed by GetOpenFileName. */
    char         szBuf[256], szPath[128], szFile[13];
    DWORD        cb;                /* Size of file. */
    DWORD        ibBase;   /* Linear memory address. */
    DWORD        ib;        /* Linear memory offset. */
    int          clin;   /* Number of lines in file. */
    WORD         sel;                   /* Selector. */
    HGLOBAL      hgbl; /* Text buffer global handle. */
    LPSTR        lpch; /* Synthesized character ptr. */
    LPLONG       lprgib;                /* Line map. */

    memset(&ofn, 0, sizeof ofn);  /* Prepare for the */
    szPath[0] = szFile[0] = '\0';       /* open file */
    ofn.lStructSize = sizeof ofn;         /* dialog. */
    ofn.hwndOwner = hwnd;
    ofn.lpstrFilter = "Text (*.txt)\0*.txt\0";
    ofn.nFilterIndex = 1L;
    ofn.lpstrFile = szPath;
    ofn.nMaxFile = sizeof szPath;
    ofn.lpstrFileTitle = szFile;
    ofn.nMaxFileTitle = sizeof szFile;
    ofn.Flags = OFN_HIDEREADONLY | OFN_FILEMUSTEXIST;
    if (!GetOpenFileName(&ofn) ||
      NULL == (hfil = _lopen(szPath, READ)))
        return;

    wsprintf(szBuf, "%s - %s",       /* Show name in */
      (LPSTR)szTitle, (LPSTR)szFile);    /* caption. */
    SetWindowText(hwnd, szBuf);

    cb = _llseek(hfil, 0, 2);      /* Get file size. */
    _llseek(hfil, 0, 0);

    /* Get space and read it into memory. */
    lpch = (LPSTR)GetWindowLong(hwnd, ibText);
    lpch = GlobalReAllocPtr(lpch, cb, GMEM_MOVEABLE);
    _hread(hfil, lpch, cb);
    _lclose(hfil);
    ((char huge *)lpch)[cb - 1] = 0;   /* Null term. */
    SetWindowLong(hwnd, ibText, (long)lpch);

    /* Allocate and prepare new selector. */
    GlobalFix(hgbl = GlobalPtrHandle(lpch));
    sel = SelAlloc(0x0000ffff);
    ibBase = IbBaseGetSel(SELECTOROF(lpch));

    /* Build up line map. */
    lprgib = (LPLONG)GetWindowLong(hwnd, ibMap);
    lprgib[0] = 0;
    clin = 1;

    /* Loop until no more lines or no more room. */
    for (ib = 0; ib < cb && clinMax > clin;
      ib++, lpch++)
        if ((*lpch == '\r' && ib + 1 < cb &&
            lpch[1] == '\n') ||
          OFFSETOF(lpch) == 0xffff)
            {
            /* New line or max line length. */
            *lpch = 0;
            lprgib[clin++] = ib + 2;

            /* Start over from current spot. */
            SetBaseSelIb(sel, ibBase + ib);
            lpch = MAKELP(sel, 0);
            }
    lprgib[clin] = -1; /* Mark end of map. */

    /* Adjust scroll bar. */
    SetScrollRange(hwnd, SB_VERT, 0, clin - 1, FALSE);
    SetScrollPos(hwnd, SB_VERT, 0, TRUE);

    FreeSel(sel);                        /* Cleanup. */
    GlobalUnfix(hgbl);
    }

void PaintText(HWND hwnd)
/*****************************************************/
/* -- Handle a paint message.                        */
/* -- Paint the text for the current scroll pos'n.   */
/*****************************************************/
    {
    WORD        sel;                    /* Selector. */
    HGLOBAL     hgbl;  /* Text buffer global handle. */
    LPSTR       lpsz;   /* Synthetic string pointer. */
    LPLONG      lpib;     /* Line map entry pointer. */
    PAINTSTRUCT wps; /* For BeginPaint()/EndPaint(). */
    RECT        rect;                /* Client area. */
    int         y, dyText;           /* Coordinates. */
    int         lin;                  /* Line index. */
    DWORD       ibBase;    /* Linear memory address. */

    BeginPaint(hwnd, &wps);
    dyText = HIWORD(GetDialogBaseUnits());

    /* Allocate and prepare selector for text. */
    lpsz = (LPSTR)GetWindowLong(hwnd, ibText);
    GlobalFix(hgbl = GlobalPtrHandle(lpsz));
    sel = SelAlloc(0x0000ffff);

    /* Get line number and map entry. */
    lpib = (LPLONG)GetWindowLong(hwnd, ibMap);
    lpib += (lin = GetScrollPos(hwnd, SB_VERT));

    /* Loop over lines in client area. */
    GetClientRect(hwnd, &rect);
    ibBase = IbBaseGetSel(SELECTOROF(lpsz));
    lpsz = MAKELP(sel, 0);
    for (y = 0; y < rect.bottom; y += dyText, lpib++)
        {
        if (lpib[1] <= -1)
            break;
        SetBaseSelIb(sel, ibBase + *lpib);
        ExtTextOut(wps.hdc, 0, y, 0, NULL, lpsz,
          lstrlen(lpsz), NULL);
        }

    FreeSel(sel);                        /* Cleanup. */
    GlobalUnfix(hgbl);
    EndPaint(hwnd, &wps);
    }

LRESULT CALLBACK __export LwWndProc(HWND hwnd, UINT wm,
  WPARAM wParam, LPARAM lParam)
/*****************************************************/
/* -- Main window procedure.                         */
/*****************************************************/
    {
    LPLONG lprgib;

    switch (wm)
        {
    default:
        break;

    case WM_CREATE: /* Allocate line map and initial */
        SetWindowLong(hwnd, ibText,  /* text buffer. */
          (LONG)GlobalAllocPtr(GMEM_MOVEABLE, cbBuf));
        lprgib = GlobalAllocPtr(GMEM_MOVEABLE, cbBuf);
        lprgib[1] = -1; /* No lines. */
        SetWindowLong(hwnd, ibMap, (LONG)lprgib);
        SetScrollRange(hwnd, SB_VERT, 0, 0, TRUE);
        break;

    case WM_DESTROY:                 /* Free memory. */
        GlobalFreePtr(GetWindowLong(hwnd, ibText));
        GlobalFreePtr(GetWindowLong(hwnd, ibMap));
        PostQuitMessage(0);
        break;

    case WM_PAINT:
        PaintText(hwnd);
        return 0;

    case WM_VSCROLL:
        ScrollText(hwnd, LOWORD(lParam), wParam);
        return 0;

    case WM_SYSCOMMAND:
        if (midOpen == wParam)
            {
            ReadText(hwnd);
            return 0;
            }
        break;
        }

    return DefWindowProc(hwnd, wm, wParam, lParam);
    }

void ScrollText(HWND hwnd, int ilin, int wType)
/*****************************************************/
/* -- Scroll the text in the window.                 */
/*****************************************************/
    {
    int  clinMin, clinLim, clinRect;
    RECT rect;

    GetScrollRange(hwnd, SB_VERT, &clinMin, &clinLim);
    if (0 == clinLim)
        return;
    GetClientRect(hwnd, &rect);
    clinRect =
      rect.bottom / HIWORD(GetDialogBaseUnits());
    if (wType != SB_THUMBPOSITION &&
      wType != SB_THUMBTRACK)
        {
        ilin = GetScrollPos(hwnd, SB_VERT);
        switch (wType)
            {
        default:
            return;

        case SB_BOTTOM:
            ilin = clinLim;
            break;

        case SB_TOP:
            ilin = 0;
            break;

        case SB_LINEDOWN:
            ilin++;
            break;

        case SB_LINEUP:
            ilin--;
            break;

        case SB_PAGEDOWN:
            ilin += clinRect;
            break;

        case SB_PAGEUP:
            ilin -= clinRect;
            break;
            }
        }

    SetScrollPos(hwnd, SB_VERT, ilin, TRUE);
    InvalidateRect(hwnd, NULL, TRUE);
    }
